@@ -1,4 +1,4 @@
-/* tftp-hpa: $Id: tftpd.c,v 1.58 2004/09/14 22:38:46 hpa Exp $ */
+/* tftp-hpa: $Id$ */
 
 /* $OpenBSD: tftpd.c,v 1.13 1999/06/23 17:01:36 deraadt Exp $	*/
 
@@ -45,7 +45,7 @@ static const char *copyright UNUSED =
 /*static char sccsid[] = "from: @(#)tftpd.c	5.13 (Berkeley) 2/26/91";*/
 /*static char rcsid[] = "$OpenBSD: tftpd.c,v 1.13 1999/06/23 17:01:36 deraadt Exp $: tftpd.c,v 1.6 1997/02/16 23:49:21 deraadt Exp $";*/
 static const char *rcsid UNUSED =
-"tftp-hpa $Id: tftpd.c,v 1.58 2004/09/14 22:38:46 hpa Exp $";
+"tftp-hpa $Id$";
 #endif /* not lint */
 
 /*
@@ -104,11 +104,12 @@ int     tsize_ok;
 int	ndirs;
 const char **dirs;
 
-int	secure = 0;
-int	cancreate = 0;
-int	unixperms = 0;
-
-int verbosity = 0;
+int		secure = 0;
+int		cancreate = 0;
+int		unixperms = 0;
+int		portrange = 0;
+unsigned int	portrange_from, portrange_to;
+int		verbosity = 0;
 
 struct formats;
 #ifdef WITH_REGEX
@@ -187,7 +188,7 @@ read_remap_rules(const char *file)
 }
 #endif
 
-static inline void
+static void
 set_socket_nonblock(int fd, int flag)
 {
   int err;
@@ -206,6 +207,16 @@ set_socket_nonblock(int fd, int flag)
   }
 }
 
+static void
+pmtu_discovery_off(int fd)
+{
+#if defined(IP_MTU_DISCOVER) && defined(IP_PMTUDISC_DONT)
+  int pmtu = IP_PMTUDISC_DONT;
+
+  setsockopt(fd, IPPROTO_IP, IP_MTU_DISCOVER, &pmtu, sizeof(pmtu));
+#endif
+}
+
 /*
  * Receive packet with synchronous timeout; timeout is adjusted
  * to account for time spent waiting.
@@ -262,6 +273,39 @@ static int recv_time(int s, void *rbuf, int len, unsigned int flags,
   }
 }
 
+static int
+pick_port_bind(int sockfd, struct sockaddr_in *myaddr)
+{
+  unsigned int port, firstport;
+
+  firstport = portrange
+    ? portrange_from + rand() % (portrange_to-portrange_from+1)
+    : 0;
+
+  port = firstport;
+
+  do {
+    myaddr->sin_port = htons(port);
+    
+    if (bind(sockfd, (struct sockaddr *)myaddr, sizeof *myaddr) < 0) {
+      /* Some versions of Linux return EINVAL instead of EADDRINUSE */
+      if ( !(portrange && (errno == EINVAL || errno == EADDRINUSE)) )
+	return -1;
+
+      /* Normally, we shouldn't have to loop, but some situations involving
+	 aborted transfers make it possible. */
+    } else {
+      return 0;
+    }
+
+    port++;
+    if ( port > portrange_to )
+      port = portrange_from;
+  } while ( port != firstport );
+
+  return -1;
+}
+
 int
 main(int argc, char **argv)
 {
@@ -293,8 +337,10 @@ main(int argc, char **argv)
   __progname = (p && p[1]) ? p+1 : argv[0];
   
   openlog(__progname, LOG_PID|LOG_NDELAY, LOG_DAEMON);
+
+  srand(time(NULL) ^ getpid());
   
-  while ((c = getopt(argc, argv, "cspvVla:B:u:U:r:t:T:m:")) != -1)
+  while ((c = getopt(argc, argv, "cspvVla:B:u:U:r:t:T:R:m:")) != -1)
     switch (c) {
     case 'c':
       cancreate = 1;
@@ -337,6 +383,16 @@ main(int argc, char **argv)
 	maxtimeout = rexmtval*TIMEOUT_LIMIT;
       }
       break;
+    case 'R':
+      {
+	if ( sscanf(optarg, "%u:%u", &portrange_from, &portrange_to) != 2 ||
+	     portrange_from > portrange_to || portrange_to >= 65535 ) {
+	  syslog(LOG_ERR, "Bad port range: %s", optarg);
+	  exit(EX_USAGE);
+	}
+	portrange = 1;
+      }
+      break;
     case 'u':
       user = optarg;
       break;
@@ -508,6 +564,9 @@ main(int argc, char **argv)
     close(1); close(2);
   }
 
+  /* Disable path MTU discovery */
+  pmtu_discovery_off(0);
+
   /* This means we don't want to wait() for children */
 #ifdef SA_NOCLDWAIT
   set_signal(SIGCHLD, SIG_IGN, SA_NOCLDSTOP|SA_NOCLDWAIT);
@@ -691,16 +750,19 @@ main(int argc, char **argv)
   from.sin_family = AF_INET;
   
   /* Process the request... */
-  
-  myaddr.sin_port = htons(0); /* We want a new local port */
-  if (bind(peer, (struct sockaddr *)&myaddr, sizeof myaddr) < 0) {
+  if (pick_port_bind(peer, &myaddr) < 0) {
     syslog(LOG_ERR, "bind: %m");
     exit(EX_IOERR);
   }
+  
   if (connect(peer, (struct sockaddr *)&from, sizeof from) < 0) {
     syslog(LOG_ERR, "connect: %m");
     exit(EX_IOERR);
   }
+
+  /* Disable path MTU discovery */
+  pmtu_discovery_off(0);
+
   tp = (struct tftphdr *)buf;
   tp_opcode = ntohs(tp->th_opcode);
   if (tp_opcode == RRQ || tp_opcode == WRQ)
